# Copyright (c) 2012, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of ingranalyze.
#
# ingranalyze is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ingranalyze is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ingranalyze.  If not, see <http://www.gnu.org/licenses/>.
# -*- coding: utf-8 -*-
from pyasp.asp import *
from pyasp.misc import *

class Lexer:
	tokens = (
		'IDENT',
		'ARROW',
		'PLUS',
		'MINUS',
		'AND',
		#'COMPL',
		#'OR',
		#'LP',
		#'RP',
		#'COM',
	)

	# Tokens

	t_IDENT = r'[a-zA-Z][a-zA-Z0-9_:\-\[\]/]*'
	t_ARROW = r'->'
	t_PLUS = r'\+'
	t_MINUS = r'-'
	t_AND = r'\?'
	t_COMPL = r'&'
	t_OR = r'\|'
	t_LP = r'\('
	t_RP = r'\)'
	t_COM = r','

	def __init__(self):
		import pyasp.ply.lex as lex
		self.lexer = lex.lex(object = self, optimize=1, lextab='graph_parser_lextab')

	# Ignored characters
	t_ignore = " \t"

	def t_newline(self, t):
		r'\n+'
		t.lexer.lineno += t.value.count("\n")
		
	def t_error(self, t):
		print "Illegal character '%s'" % t.value[0]
		t.lexer.skip(1)


class Parser:
	tokens = Lexer.tokens

	precedence = ( )
		#('left','PLUS','MINUS'),
		#('left','TIMES','DIVIDE'),
		#('right','UMINUS'),
		#)

	def __init__(self):
		self.aux_node_counter=0
		self.accu = TermSet()
		self.fun_flag = False
		self.args = []
		self.commas = []
		self.lexer = Lexer()
		import pyasp.ply.yacc as yacc
		#self.parser = yacc.yacc(module=self, debug=False, tabmodule='calc_parsetab', debugfile="calc_parser.out")
		self.parser = yacc.yacc(module=self, optimize=1, tabmodule='graph_parser_parsetab',)

	def p_statement_expr(self, t):
		'''statement : node_expression ARROW node_expression value
					| node_expression '''
		if len(t)<3 : 
			self.accu.add(Term('input', [t[1]]))
		else :
			self.accu.add(Term('edge', [t[1],t[3]]))
			if t[4]!="0" : self.accu.add(Term('obs_elabel', [t[1],t[3],t[4]]))
			#self.names[t[0]] = t[2]

	def p_node_expression(self, t):
		'''node_expression : IDENT '''

		if len(t)<3 : 
			
			if self.fun_flag==True :
				t[0]=t[1]
				self.fun_flag=False
			else :
				t[0] = "gen(\""+t[1]+"\")"
				self.accu.add(Term('vertex', ["gen(\""+t[1]+"\")"]))
		elif t[1] == 'opposite_sign' :
			self.aux_node_counter+=1
			aux_node = "aux_node_"+str(self.aux_node_counter)
			t[0] = "opposite_sign"+"("+aux_node+")"
			self.accu.add(Term('edge', [t[3],t[1]+"("+aux_node+")"]))
			self.accu.add(Term('obs_elabel', [t[3],t[1]+"("+aux_node+")","1"]))
		elif t[1] == 'strong_inhibitor' :
			t[0] = "strong_inhibitor"+"("+t[3]+")"
			for i in self.commas : 
				self.accu.add(Term('edge', [i,t[1]+"("+t[3]+")"]))
				self.accu.add(Term('obs_elabel', [i,t[1]+"("+t[3]+")","1"]))
			self.commas= []
		else : t[0] = "unknown"

	def p_value(self, t):
		'''value : PLUS
		       | MINUS
		       | AND '''
		if t[1] == '-' : t[0] = "-1"
		elif t[1] == '+' : t[0] = "1"
		elif t[1] == '?' : t[0] = "0"

	#def p_commalist(self, t):
		#'''commalist : IDENT COM IDENT'''
					  ##| commalist COM IDENT'''
		##t[0] = t[1]+" , "+t[3]
		#t[0] = "gen(\""+t[1]+"\"),gen(\""+t[3]+"\")"
		#self.accu.add(Term('vertex', ["gen(\""+t[1]+"\")"]))
		#self.accu.add(Term('vertex', ["gen(\""+t[3]+"\")"]))
		#self.commas.append("gen(\""+t[1]+"\")")
		#self.commas.append("gen(\""+t[3]+"\")")

	#def p_orexpression2(self, t):
		#'orlist2 : orlist'
		#self.aux_node_counter+=1
		#aux_node = "aux_node_"+str(self.aux_node_counter)
		#self.accu.add(Term('vertex', ["or("+aux_node+")"]))
		#for i in self.args : 
			#self.accu.add(Term('edge', [i,"or("+aux_node+")"]))
			#self.accu.add(Term('obs_elabel', [i,"or("+aux_node+")","1"]))
		#self.args= []
		#t[0]="or("+aux_node+")"
		
	#def p_andexpression2(self, t):
		#'andlist2 : andlist'
		#self.aux_node_counter+=1
		#aux_node = "aux_node_"+str(self.aux_node_counter)
		#self.accu.add(Term('vertex', ["and("+aux_node+")"]))
		#for i in self.args : 
			#self.accu.add(Term('edge', [i,"and("+aux_node+")"]))
			#self.accu.add(Term('obs_elabel', [i,"and("+aux_node+")","1"]))
		#self.args= []
		#t[0]="and("+aux_node+")"

	#def p_orexpression(self, t):
		#'''orlist : IDENT
				#| orlist OR IDENT'''
		#self.fun_flag=True
		#if len(t)<3 : 
			#self.accu.add(Term('vertex', ["gen(\""+t[1]+"\")"]))
			#self.args.append("gen(\""+t[1]+"\")")
			#t[0] = "gen(\""+t[1]+"\")"
		#elif t[2] == '|' : 
			#self.accu.add(Term('vertex', ["gen(\""+t[3]+"\")"]))
			#self.args.append("gen(\""+t[3]+"\")")
			#t[0] = "Unknown"

	#def p_andexpression(self, t):
		#'''andlist : IDENT
					  #| andlist COMPL IDENT'''
		#self.fun_flag=True
		#if len(t)<3 : 
			#self.accu.add(Term('vertex', ["gen(\""+t[1]+"\")"]))
			#self.args.append("gen(\""+t[1]+"\")")
			#t[0] = "gen(\""+t[1]+"\")"
		#elif t[2] == '&' : 
			#self.accu.add(Term('vertex', ["gen(\""+t[3]+"\")"]))
			#self.args.append("gen(\""+t[3]+"\")")
			#t[0] = "Unknown"
			
	def p_error(self, t):
		print "Syntax error at '%s'" % t

	def parse(self, line):
		self.parser.parse(line, lexer=self.lexer.lexer)
		return self.accu



# Copyright (c) 2012, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of ingranalyze.
#
# ingranalyze is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ingranalyze is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ingranalyze.  If not, see <http://www.gnu.org/licenses/>.
# -*- coding: utf-8 -*-
import os

#some pretty printers for predictions, minimal inconsistent cores, etc

def print_predictions(predictions) :
    #predictions.sort()
    predictions = sorted(predictions, key=lambda p: str(p.arg(0)))
    exp = ''
    for p in predictions:
        if p.pred() == "vlabel" :
           if exp!=str(p.arg(0)) :
              print 'Experiment '+str(p.arg(0))+':'
              exp=str(p.arg(0))
           if p.arg(2)== "-1": print('   '+str(p.arg(1))+ ' = - ')
           if p.arg(2)== "1" : print('   '+str(p.arg(1))+ ' = + ')
           if p.arg(2)== "0" : print('   '+str(p.arg(1))+ ' = nc ')
        if p.pred() == "elabel" :
           if p.arg(2) == "-1" : print '   '+str(p.arg(0))+' -> '+str(p.arg(1))+' -'
           if p.arg(2) == "1"  : print '   '+str(p.arg(0))+' -> '+str(p.arg(1))+' +'
                  
def print_mic(mic, net, obs):
  
    nodes = []
    edges = []
    for node in mic: nodes.append(str(node.arg(1)))
    
    predecessors = []
    for e in net:
       if e.pred() == "obs_elabel" :
          #print str(e)
          #print str(e.arg(0)),str(e.arg(1)),str(e.arg(2))
          if str(e.arg(1)) in nodes : 
            predecessors.append(str(e.arg(0)))
            if str(e.arg(2)) == "1" : edges.append( str(e.arg(0))+ " -> " + str(e.arg(1))+ " +")
            if str(e.arg(2)) == "-1" : edges.append(str(e.arg(0))+ " -> " + str(e.arg(1))+ " -")
         #TODO ? edges
    for edge in edges: print('   '+edge)
    for o in obs:
       if o.pred() == "obs_vlabel" :  
          if str(o.arg(1)) in nodes :
              if str(o.arg(2))=="1" :  print '   '+str(o.arg(1))+ " = +"
              if str(o.arg(2))=="-1" :  print '   '+str(o.arg(1))+ " = -"
          if str(o.arg(1)) in predecessors :
              if str(o.arg(2))=="1" :  print '   '+str(o.arg(1))+ " = +"
              if str(o.arg(2))=="-1" :  print '   '+str(o.arg(1))+ " = -"
    

def clean_up() :
  if os.path.isfile("parser.out"): os.remove("parser.out")
  if os.path.isfile("asp_py_lextab.py"): os.remove("asp_py_lextab.py")
  if os.path.isfile("asp_py_lextab.pyc"): os.remove("asp_py_lextab.pyc")
  if os.path.isfile("asp_py_parsetab.py"): os.remove("asp_py_parsetab.py")
  if os.path.isfile("asp_py_parsetab.pyc"): os.remove("asp_py_parsetab.pyc") 
  if os.path.isfile("graph_parser_lextab.py"): os.remove("graph_parser_lextab.py")
  if os.path.isfile("graph_parser_parsetab.py"): os.remove("graph_parser_parsetab.py")

def m_quit() :  
  clean_up()
  quit()


# Copyright (c) 2012, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of ingranalyze.
#
# ingranalyze is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ingranalyze is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ingranalyze.  If not, see <http://www.gnu.org/licenses/>.
# -*- coding: utf-8 -*-
import re
from pyasp.asp import *
from pyasp.misc import *

import pyasp.ply.lex as lex
import pyasp.ply.yacc as yacc

import graph_parser
import sif_parser
import profile_parser


def int_of_sign(s):
    if s == '+': return '1'
    elif s == '-': return '-1'
    elif s == '0': return '0'
    elif s == 'nc': return '0'
    else: 
        print s
        assert(False)

def sign_of_int(s):
    if s == '1': return '+'
    elif s == '-1': return '-'
    elif s == '0': return '0'
    else: 
        print s
        assert(False)



GENE_ID = '[-a-zA-Z0-9_:&\(\)/]+'

#tokens = ( 'COMPLEX', 'INFL', 'SIGN', 'GENE',)
#t_GENE = r'[a-zA-Z_][a-zA-Z0-9_]*' 
#operator = ['&','|']

def readGraph(filename):
    p = graph_parser.Parser()
    """
    input: string, name of a file containing a Bioquali-like graph description
    output: asp.TermSet, with atoms matching the contents of the input file
    
    Parses a Bioquali-like graph description, and returns
    a TermSet object.
    Written using original Bioquali
    """
	
    accu = TermSet()
    file = open(filename,'r')
    s = file.readline()
    while s!="":
        try:
            accu = p.parse(s)
        except EOFError:
            break
        s = file.readline()

    return accu

def readSIFGraph(filename):
    p = sif_parser.Parser()
    """
    input: string, name of a file containing a Bioquali-like graph description
    output: asp.TermSet, with atoms matching the contents of the input file
    
    Parses a Bioquali-like graph description, and returns
    a TermSet object.
    Written using original Bioquali
    """
	
    accu = TermSet()
    file = open(filename,'r')
    s = file.readline()
    while s!="":
        try:
            accu = p.parse(s)
        except EOFError:
            break
        s = file.readline()

    return accu


def saveGraph(termset,filename):
    file = open(filename, 'w')
    elabels = termset.filter(lambda t: t.sip('obs_elabel'))
    edges = termset.filter(lambda t: t.sip('edge'))
    inputs = termset.filter(lambda t: t.sip('input'))
    for t in elabels:
        [ _, j, i, s ] = t.explode()
        file.write(unquote(j) + ' -> ' + unquote(i) + ' ' + sign_of_int(s) + '\n')
        edges.discard(Term('edge',[j, i]))
    for t in edges:
        [_, j, i] = t.explode()
        file.write(unquote(j) + ' -> ' + unquote(i) + ' X\n')
    for t in inputs:
        [ _, i] = t.explode()
        file.write(unquote(i) + '\n')
    file.close()


def readProfile_new(filename):
    p = profile_parser.Parser()
    """
    input: string, name of a file containing a Bioquali-like graph description
    output: asp.TermSet, with atoms matching the contents of the input file
    
    Parses a Bioquali-like graph description, and returns
    a TermSet object.
    Written using original Bioquali
    """
	
    accu = TermSet()
    file = open(filename,'r')
    s = file.readline()
    while s!="":
        if s!="\n":
	    try:
		accu = p.parse(s)
	    except EOFError:
		break
	s = file.readline()

    return accu
    
def readProfile(filename):
    #COMPLEX_ID = '[-a-zA-Z0-9_:\(\)/&]+'
    GENE_ID = '[-a-zA-Z0-9_:\(\)/]+'
    SIGN = '[-+0n]c*'
    file = open(filename,'r')
    val_re = '(?P<genid>'+GENE_ID+')( )*=( )*(?P<sign>'+SIGN+')'
    val = re.compile(val_re)
    name_re = '(-\*-)( )*(?P<name>\S+)( )*(-\*-)'
    name = re.compile(name_re)
    obs_name = None
    line_number = 1
    line = file.readline()[:-1]
    # the first line may contain the name
    nm = name.match(line)
    if nm:
        name = nm.group('name')
        line = file.readline()[:-1]
    else:
        name = filename
    name = quote(name)
    accu = TermSet()
    accu.add(Term('exp',[name]))
    # ordinary line
    while line:
        vm = val.match(line)
        if vm:
            vertex = quote(vm.group('genid'))
            accu.add(Term('obs_vlabel',[name, "gen("+vertex+")", int_of_sign(vm.group('sign'))]))
        else:
            print 'Syntax error line:', line_number,'  '+line 
        line = file.readline()[:-1]
        line_number+=1
    return accu

def readProfile_with_nc(filename):
    #COMPLEX_ID = '[-a-zA-Z0-9_:\(\)/&]+'
    GENE_ID = '[-a-zA-Z0-9_:\(\)/]+'
    file = open(filename,'r')
    val_re = '(?P<genid>'+GENE_ID+')( )*=( )*(?P<sign>[-+0n]c*)( )*(?P<input>\(input\))*'
    val = re.compile(val_re)
    name_re = '(-\*-)( )*(?P<name>\S+)( )*(-\*-)'
    name = re.compile(name_re)
    obs_name = None
    line_number = 1
    line = file.readline()[:-1]
    # the first line may contain the name
    nm = name.match(line)
    if nm:
        name = nm.group('name')
        line = file.readline()[:-1]
    else:
        name = filename
    name = quote(name)
    accu = TermSet()
    accu.add(Term('exp',[name]))
    # ordinary line
    while line:
        vm = val.match(line)
        if vm:
            vertex = quote(vm.group('genid'))

            accu.add(Term('obs_vlabel',[name, "gen("+vertex+")", int_of_sign(vm.group('sign'))]))
            if vm.group('input')!=None : 
               accu.add(Term('input',[name, "gen("+vertex+")"]))
        else:
            print 'Syntax error line:', line_number,'  '+line 
        line = file.readline()[:-1]
        line_number+=1
    return accu
    
def saveProfile(termset,condition,filename):
    file = open(filename, 'w')
    file.write('-*- ' + condition + ' -*-\n')
    for t in termset:
        if t.sip('obs_vlabel'):
            [_, c, i, s] = t.explode()
            if condition == unquote(c):
                file.write(unquote(i) + ' = ' + sign_of_int(s) + '\n')
    file.close()

                



# Copyright (c) 2012, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of ingranalyze.
#
# ingranalyze is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ingranalyze is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ingranalyze.  If not, see <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
from pyasp.asp import *
from pyasp.misc import *

class Lexer:
	tokens = (
		'IDENT',
		'ASSIGN',
		'PLUS',
		'MINUS',
		'COMMENT',
	)

	# Tokens
	t_IDENT = r'[a-zA-Z][a-zA-Z0-9_:\-\[\]/]*'
	t_ASSIGN = r'='
	t_PLUS = r'\+'
	t_MINUS = r'-'
	t_COMMENT = r'%+'


	def __init__(self):
		import bioasp.ply.lex as lex
		self.lexer = lex.lex(object = self, optimize=1, lextab='profile_parser_lextab')

	# Ignored characters
	t_ignore = " \t"

	def t_newline(self, t):
		r'\n+'
		t.lexer.lineno += t.value.count("\n")
		
	def t_error(self, t):
		print "Illegal character '%s'" % t.value[0]
		t.lexer.skip(1)


class Parser:
	tokens = Lexer.tokens

	precedence = ( )
		#('left','PLUS','MINUS'),
		#('left','TIMES','DIVIDE'),
		#('right','UMINUS'),
		#)

	def __init__(self):
		self.experiment_name=""
		self.accu = TermSet()
		self.lexer = Lexer()
		import bioasp.ply.yacc as yacc
		#self.parser = yacc.yacc(module=self, tabmodule='calc_parsetab', debugfile="calc_parser.out")
		self.parser = yacc.yacc(module=self, optimize=1)

	def p_statement_expr(self, t):
		'''statement : IDENT ASSIGN value
					| IDENT 
					| COMMENT IDENT'''
		if len(t)==2 : 
			self.accu.add(Term('input', [ self.experiment_name, "gen(\""+t[1]+"\")"]))			
		if len(t)==3 : 
			self.experiment_name = "\""+t[2]+"\""
		if len(t)==4  :
			self.accu.add(Term('obs_vlabel', [self.experiment_name, "gen(\""+t[1]+"\")", t[3]]))
			


	def p_value(self, t):
		'''value : PLUS
		       | MINUS 
		       | IDENT '''
		if t[1] == '-' : t[0] = "-1"
		elif t[1] == '+' : t[0] = "1"
                elif t[1] == 'nc': t[0] = "0"

			
	def p_error(self, t):
		print "Syntax error at '%s'" % t

	def parse(self, line):
		self.parser.parse(line, lexer=self.lexer.lexer)
		return self.accu



#!python
# Copyright (c) 2012, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of ingranalyze.
#
# ingranalyze is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ingranalyze is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ingranalyze.  If not, see <http://www.gnu.org/licenses/>.
# -*- coding: utf-8 -*-
import argparse
from pyasp.asp import *
from __ingranalyze__ import query, utils, bioquali

if __name__ == '__main__':

        
    parser = argparse.ArgumentParser()
    parser.add_argument("networkfile",
                        help="influence graph in bioquali format")
    parser.add_argument("observationfile",
                        help="observations in bioquali format") 
    
    parser.add_argument('--mics', 
			help="compute minimal inconsistent cores",
			action="store_true")
			
    parser.add_argument('--repair', type=int, choices=[1, 2, 3, 4, 5,], default=3,
			help="choose repair method: 1 flip observed variations, 2 flip influences, 3 define network nodes as inputs,  4 define network nodes as input in an experiment (use only in case of multiple experiments), 5 add influences. default is 3")
			
    parser.add_argument('--list_repairs', 
			help="compute all minimal repair sets",
			action="store_true")
    
    args = parser.parse_args()
        

    net_string = args.networkfile
    obs_string = args.observationfile
    
    print '\nReading network',net_string, '...',
    net = bioquali.readGraph(net_string)
    print 'done.'
    #net.to_file("net.lp")

    print '\nReading observations',obs_string, '...',
    mu = bioquali.readProfile(obs_string)
    print 'done.'
    print '\nComputing input nodes ...',
    inputs = query.guess_inputs(net)
    print 'done.'

    print '\nTesting empty network for consistency ...',
    consistent = query.is_consistent(net)
    print 'done.'

    if consistent: print '   The empty network is consistent.'
    else:
      print '   The empty network is inconsistent.'

      empty_net = net.union(inputs)
      print '\nTesting empty network plus input nodes for consistency ...',
      consistent = query.is_consistent(empty_net)
      print 'done.'
      if consistent: print '   The empty network is consistent.'
      else:
	print '   The empty network is still inconsistent.'
	
	if args.mics: 
	  print '\nComputing minimal inconsistent cores (mic\'s) ...',
	  mics = query.get_minimal_inconsistent_cores(empty_net)
	  print 'done.\n'
	  count = 1
	  oldmic = 0
	  for mic in mics:
	      if oldmic != mic: 
		print 'mic '+str(count)+':'
		utils.print_mic(mic.to_list(),net.to_list(),[])
		count+=1
		oldmic= mic

	repair_options= TermSet()
	print '\nCompute repair options ...',
	if args.repair==1:                    
	  print 'repair mode: flip observed variations ...',
	  repair_options = query.get_repair_options_flip_obs(empty_net)
	  print 'done.'
	if args.repair==2:                    
	  print 'repair mode: flip influences ...',
	  repair_options = query.get_repair_options_flip_edge(empty_net)
	  print 'done.'
	if args.repair==3:                    
	  print 'repair mode: define network nodes as inputs ...',
	  repair_options = query.get_repair_options_make_node_input(empty_net)
	  print 'done.'
	if args.repair==4:
	  print 'repair mode: define network nodes as input in an experiment ...',
	  repair_options = query.get_repair_options_make_obs_input(empty_net)
	  print 'done.'                    
	if args.repair==5:
	  print 'repair mode: add influence ...',
	  repair_options = query.get_repair_options_add_edges(empty_net)
	  print 'done.'                    
      

	print '\nCompute minimal numbers of necessary repair operations ...',
	optimum = query.get_minimum_of_repairs(empty_net,repair_options) 
	print 'done.'    
	
	print '   The data set can be repaired with minimal', optimum[0],'operations.'
	
	do_repair= raw_input('\nDo you want to compute all possible repair sets? Y/n:')
	if do_repair=="Y":

	  print '\nComputing all repair sets with size', optimum[0],'...'
	  models = query.get_minimal_repair_sets(empty_net,repair_options,optimum[0])
	  print 'done.'
	
	  count = 1
	  oldmodel = 0
	  for model in models:
	    if oldmodel != model: 
	      oldmodel = model
	      repairs = model.to_list()
	      print '  repair',count,':'
	      for r in repairs : print str(r.arg(0)),
	      print ' '
	      count+=1
	
	print '\nComputing predictions that hold under all repair sets size', optimum[0],'...',
	model = query.get_predictions_under_minimal_repair(optimum[0],empty_net,repair_options)
	print 'done.'
	predictions = model.to_list()
	print ( str(len(predictions)) + ' predictions found:')
	utils.print_predictions(predictions)

    if consistent: #if network is consistent add data
      net_with_data = net.union(mu).union(inputs)
      print '\nTesting network with data for consistency ...',
      consistent = query.is_consistent(net_with_data)
      print "done."
      if consistent: 
	print '   The network and data are consistent.'
	
	print '\nComputing predictions under consistency ...',
	model = query.get_predictions_under_consistency(net_with_data)
	print 'done.'
	predictions = model.to_list()
	#predictions.sort()
	print ( str(len(predictions)) + ' predictions found:')
	utils.print_predictions(predictions)
	
	
      else:
	print '   The network and the data are inconsistent.'
	
	if args.mics: 
	  print '\nComputing minimal inconsistent cores (mic\'s) ...',
	  mics = query.get_minimal_inconsistent_cores(net_with_data)
	  print 'done.'
	  count = 1
	  oldmic = 0    
	  for mic in mics:
	    if oldmic != mic: 
	      print 'mic '+str(count)+':'
	      utils.print_mic(mic.to_list(),net.to_list(),mu.to_list())
	      count+=1
	      oldmic= mic
	    

	repair_options= TermSet()
	print '\nCompute repair options ...',
	if args.repair==1:
	  print 'repair mode: flip observed variations ...',
	  repair_options = query.get_repair_options_flip_obs(net_with_data)
	  print 'done.'
	if args.repair==2:                    
	  print 'repair mode: flip influences ...',    
	  repair_options = query.get_repair_options_flip_edge(net_with_data)
	  print 'done.'
	if args.repair==3:                    
	  print 'repair mode: define network nodes as inputs ...',
	  repair_options = query.get_repair_options_make_node_input(net_with_data)
	  print 'done.'
	if args.repair==4:                    
	  print 'repair mode: define network nodes as input in an experiment ...',
	  repair_options = query.get_repair_options_make_obs_input(net_with_data)
	  print 'done.'
	if args.repair==5:                    
	  print 'repair mode: add influence ...',
	  repair_options = query.get_repair_options_add_edges(net_with_data)
	  print 'done.'
      

	print '\nCompute minimal numbers of necessary repair operations ...',
	optimum = query.get_minimum_of_repairs(net_with_data, repair_options)
	print 'done.'
	    
	print '   The data set can be repaired with minimal', optimum.score[0],'operations.'
	  
	if args.list_repairs:

	  print '\nComputing all repair sets with size', optimum.score[0],'...',
	  models = query.get_minimal_repair_sets(net_with_data, repair_options, optimum.score[0])
	  print "done."
	  count = 1
	  oldmodel = 0
	  for model in models:
	    if oldmodel != model: 
	      oldmodel = model
	      repairs = model.to_list()
	      print "  repair",count,':'
	      for r in repairs : print str(r.arg(0)),
	      print ' '
	      count+=1

	#print 'Computing subset minimal repairs ...'
	#print '\n', query.subset_minimal_repair_flip_obs(net_with_data)
	
	print '\nComputing predictions that hold under all repair sets size', optimum.score[0],'...',
	model = query.get_predictions_under_minimal_repair(net_with_data, repair_options, optimum.score[0])
	print "done."
	predictions = model.to_list()
	#predictions.sort()
	print ( str(len(predictions)) + ' predictions found:')
	utils.print_predictions(predictions)

    utils.clean_up()  



# Copyright (c) 2012, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of ingranalyze.
#
# ingranalyze is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ingranalyze is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ingranalyze.  If not, see <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
import os
import tempfile
from pyasp.asp import *

root = __file__.rsplit('/', 1)[0]

guess_inputs_prg = root + '/encodings/pm_guess_inputs.gringo'

consistency_prg = root + '/encodings/pm_consistency.gringo'
prediction_prg = root + '/encodings/pm_prediction.gringo'
explanation_prg = root + '/encodings/explanation.gringo'
mic_prg = root + '/encodings/pm_mic.gringo'
dyn_mic_prg = root + '/encodings/dynamic_diagnosis.gringo'
reduction_prg = root + '/encodings/reduction.lp'

repair_options_prg =     root + '/encodings/compute_repair_options.gringo'
repair_core_prg =        root + '/encodings/repair_core.gringo'
prediction_core_prg =    root + '/encodings/prediction_core.gringo'
repair_cardinality_prg = root + '/encodings/repair_cardinality.gringo'
repair_subset_prg  =     root + '/encodings/repair_subset.gringo'

def is_consistent(instance):
    '''
    [is_consistent(instance)] returns [True] if there exists a consistent extension
    to the system described by the TermSet object [instance].
    '''
    return get_consistent_labelings(instance,1) != []
    

def get_consistent_labelings(instance,nmodels=0,exclude=[]):
    '''
    [consistent_labelings(instance,nmodels,exclude)] returns a list containing
    [nmodels] TermSet objects representing consistent extensions of the system
    described by the TermSet [instance]. The list [exclude] should contain TermSet objects
    representing (maybe partial) solutions that are to be avoided. If [nmodels] equals [0]
    then the list contain all feasible models.
    '''
    inputs = get_reductions(instance)
    prg = [ consistency_prg, inputs.to_file(), instance.to_file(), exclude_sol(exclude) ]
    solver = GringoClasp()
    models = solver.run(prg,nmodels)
    os.unlink(prg[1])
    os.unlink(prg[2])
    os.unlink(prg[3])
    return models
   
    
#def get_minimal_inconsistent_cores(instance,nmodels=0,exclude=[]):
    #'''
    #[compute_mic(instance,nmodels,exclude)] returns a list containing
    #[nmodels] TermSet objects representing subset minimal inconsistent cores of the system
    #described by the TermSet [instance]. The list [exclude] should contain TermSet objects
    #representing (maybe partial) solutions that are to be avoided. If [nmodels] equals [0]
    #then the list contain all feasible models.
    #'''
    #inputs = get_reductions(instance)
    #prg = [ mic_prg, inputs.to_file(), instance.to_file(), exclude_sol(exclude) ] 
    #options=' --heuristic=Vmtf'
    #solver = GringoClaspD(options)
    #models = solver.run(prg,nmodels)
    #os.unlink(prg[1])
    #os.unlink(prg[2])
    #os.unlink(prg[3])
    #return models[0]

def get_minimal_inconsistent_cores(instance,nmodels=0,exclude=[]):
    '''
    [compute_mic(instance,nmodels,exclude)] returns a list containing
    [nmodels] TermSet objects representing subset minimal inconsistent cores of the system
    described by the TermSet [instance]. The list [exclude] should contain TermSet objects
    representing (maybe partial) solutions that are to be avoided. If [nmodels] equals [0]
    then the list contain all feasible models.
    '''
    inputs = get_reductions(instance)
    prg = [ dyn_mic_prg, inputs.to_file(), instance.to_file(), exclude_sol(exclude) ] 
    options='--heuristic=Vmtf'
    solver = GringoClasp(clasp_options=options)
    models = solver.run(prg,nmodels=0,collapseTerms=True, collapseAtoms=False)
    os.unlink(prg[1])
    os.unlink(prg[2])
    os.unlink(prg[3])
    return models

def guess_inputs(instance):
    prg = [ guess_inputs_prg, instance.to_file() ]
    solver = GringoClasp()
    models = solver.run(prg,0, collapseTerms=True, collapseAtoms=False)
    os.unlink(prg[1])
    assert(len(models) == 1)
    return models[0]

def get_reductions(instance):
    prg = [ reduction_prg, instance.to_file() ]
    solver = GringoClasp()
    models = solver.run(prg,0)
    os.unlink(prg[1])
    assert(len(models) == 1)
    return models[0]


def get_repair_options_flip_obs(instance):
    repair_mode = String2TermSet('repair_v')
    instance2 = instance.union(repair_mode)
    prg = [ instance2.to_file(), repair_options_prg ]
    solver = GringoClasp()
    models = solver.run(prg)
    os.unlink(prg[0])    
    return models[0]
    

def get_repair_options_flip_edge(instance):
    repair_mode = String2TermSet('repair_e')
    instance2 = instance.union(repair_mode)
    prg = [ instance2.to_file(), repair_options_prg ]
    solver = GringoClasp()
    models = solver.run(prg)
    os.unlink(prg[0])    
    return models[0]

def get_repair_options_make_node_input(instance):
    repair_mode = String2TermSet('repair_g')
    instance2 = instance.union(repair_mode)
    prg = [ instance2.to_file(), repair_options_prg ]
    solver = GringoClasp()
    models = solver.run(prg)
    os.unlink(prg[0])    
    return models[0]
    
      
def get_repair_options_make_obs_input(instance):
    repair_mode = String2TermSet('repair_i')
    instance2 = instance.union(repair_mode)
    prg = [ instance2.to_file(), repair_options_prg ]
    solver = GringoClasp()
    models = solver.run(prg,1)
    os.unlink(prg[0])    
    return models[0]
    
def get_repair_options_add_edges(instance):
    repair_mode = String2TermSet('repair_a')
    instance2 = instance.union(repair_mode)
    prg = [ instance2.to_file(), repair_options_prg ]
    solver = GringoClasp()
    models = solver.run(prg)
    os.unlink(prg[0])    
    return models[0]


def get_minimum_of_repairs(instance,repair_options,exclude=[]):
    inputs = get_reductions(instance)
    instance2 = instance.union(inputs)
    prg = [ instance2.to_file(),repair_options.to_file(), exclude_sol(exclude), repair_core_prg, repair_cardinality_prg ]

    solver = GringoClasp()
    optimum = solver.run(prg,nmodels=0)
    os.unlink(prg[0])
    os.unlink(prg[1])
    os.unlink(prg[2]) 
    return optimum[0]


def get_minimal_repair_sets(instance, repair_options ,optimum,nmodels=0,exclude=[]):
    inputs = get_reductions(instance)
    instance2 = instance.union(inputs)
    prg = [ instance2.to_file(), repair_options.to_file(), exclude_sol(exclude), repair_core_prg, repair_cardinality_prg ]

    options='--project --opt-mode=optN'
    solver = GringoClasp(clasp_options=options)
    models = solver.run(prg,nmodels=0, collapseTerms=True, collapseAtoms=False)
    os.unlink(prg[0])
    os.unlink(prg[1])
    os.unlink(prg[2])
    return models
    
     
def get_predictions_under_minimal_repair(instance, repair_options, optimum):
    '''
    Computes the set of signs on edges/vertices that can be cautiously
    derived from [instance], minus those that are a direct consequence
    of obs_[ev]label predicates
    '''
    inputs = get_reductions(instance)
    instance2 = instance.union(inputs)
    
    prg = [ instance2.to_file(), repair_options.to_file(), prediction_core_prg, repair_cardinality_prg ]

    options='--project --enum-mode cautious --opt-mode=optN'
    solver = GringoClasp(clasp_options=options)
    models = solver.run(prg,nmodels=0,collapseTerms=True, collapseAtoms=False)
    os.unlink(prg[0])
    os.unlink(prg[1])
    return whatsnew(instance,models[0])


def cut_obs_(s):
    return str(s)[4:]

def whatsnew(instance,pred):
    '''
    [whatsnew(instance,pred)] is a TermSet equal to [pred] where all predicates
    vlabel and elabel which have a corresponding obs_vlabel and obs_elabel in
    [instance] have been deleted. This function is meant to see which of the invariants
    are not a direct consequence of the observations.
    '''
    accu = TermSet(pred)
    for t in instance:

        if t.pred() == 'obs_vlabel':
            [_,e,v,s] = t.explode()
            accu.discard(Term('vlabel',[e,v,s]))
        elif t.p('obs_elabel'):
            [_,j,i,s] = t.explode()
            accu.discard(Term('elabel',[j,i,s]))
    return accu


def get_predictions_under_consistency(instance):
    '''
    Computes the set of signs on edges/vertices that can be cautiously
    derived from [instance], minus those that are a direct consequence
    of obs_[ev]label predicates
    '''
    inputs = get_reductions(instance)

    prg = [ prediction_prg, inputs.to_file(), instance.to_file(), exclude_sol([]) ]
    solver = GringoClasp(clasp_options='--project --enum-mode cautious')
    models = solver.run(prg,0,collapseTerms=True, collapseAtoms=False)
    os.unlink(prg[1])
    os.unlink(prg[2])
    os.unlink(prg[3])
    return whatsnew(instance,models[0])
    


# Copyright (c) 2012, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of ingranalyze.
#
# ingranalyze is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ingranalyze is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ingranalyze.  If not, see <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-

from pyasp.asp import *
from pyasp.misc import *
class Lexer:
	tokens = (
		'IDENT',
		'PLUS',
		'MINUS',
	)

	# Tokens

	t_IDENT = r'[a-zA-Z][a-zA-Z0-9_:\-\[\]/]*'

	t_PLUS = r'1'
	t_MINUS = r'-1'


	def __init__(self):
		import bioasp.ply.lex as lex
		self.lexer = lex.lex(object = self, optimize=1, lextab='sif_parser_lextab')

	# Ignored characters
	t_ignore = " \t"

	def t_newline(self, t):
		r'\n+'
		t.lexer.lineno += t.value.count("\n")
		
	def t_error(self, t):
		print "Illegal character '%s'" % t.value[0]
		t.lexer.skip(1)


class Parser:
	tokens = Lexer.tokens

	precedence = ( )

	def __init__(self):
		# dictionary of names
		self.aux_node_counter=0
		#self.names = { }
		self.accu = TermSet()
		self.args = []
		self.lexer = Lexer()
		import bioasp.ply.yacc as yacc
		#self.parser = yacc.yacc(module=self, tabmodule='calc_parsetab', debugfile="calc_parser.out")
		self.parser = yacc.yacc(module=self,optimize=1)

	def p_statement_expr(self, t):
		'''statement : node_expression PLUS node_expression 
					| node_expression MINUS node_expression'''
		if len(t)<3 : 
			self.accu.add(Term('input', [t[1]]))
			print 'input', t[1]
		else :
			#print t[1], t[2], t[3]
			self.accu.add(Term('edge', ["gen(\""+t[1]+"\")","gen(\""+t[3]+"\")"]))
			self.accu.add(Term('obs_elabel', ["gen(\""+t[1]+"\")","gen(\""+t[3]+"\")",t[2]]))
			#print Term('obs_elabel', ["gen(\""+t[1]+"\")","gen(\""+t[3]+"\")",t[2]])


	def p_node_expression(self, t):
		'''node_expression : IDENT'''
		if len(t)<3 : 
				t[0]=t[1]
				#print t[1]
				self.accu.add(Term('vertex', ["gen(\""+t[1]+"\")"]))
		else : t[0] = "unknown"

			
	def p_error(self, t):
		print "Syntax error at '%s'" % t

	def parse(self, line):
		self.parser.parse(line, lexer=self.lexer.lexer)
		return self.accu



# Copyright (c) 2012, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of ingranalyze.
#
# ingranalyze is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ingranalyze is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ingranalyze.  If not, see <http://www.gnu.org/licenses/>.
# -*- coding: utf-8 -*-
from setuptools import setup
import os
import sys
import platform
import distutils
import site
import sysconfig

from setuptools.command.install import install as _install



class install(_install):
    def run(self):
        _install.run(self)
                         
setup(cmdclass={'install': install},
      name='ingranalyze',
      version='1.5',
      url='http://pypi.python.org/pypi/ingranalyze/',
      license='GPLv3+',
      description='Influence graph analysis, consistency check, diagnosis, repair and prediction ',
      long_description=open('README').read(),
      author='Sven Thiele',
      author_email='sthiele78@gmail.com',
      packages = ['__ingranalyze__'],
      package_dir = {'__ingranalyze__' : 'src'},
      package_data = {'__ingranalyze__' : ['encodings/*.lp','encodings/*.gringo']},
      scripts = ['ingranalyze.py'],
      install_requires=[
        "pyasp == 1.3"
      ]
)
